#include "bv.h"
#include "kernel.h"
#include "vector.h"
#include "object.h"
#include <math.h>

Vector BV::normals[NUM_PSN]=
{// from glassners "space-time" article
Vector(1,0,0),
Vector(0,1,0),
Vector(0,0,1),
Vector(0,0,0),
Vector(0.5, 0.5, 0.5),
Vector(0.5, 0.5, 0.5),
Vector(0.5, 0.5,-0.5),
Vector(0.5, 0.5,-0.5),

Vector(0.5,-0.5, 0.5),
Vector(0.5,-0.5, 0.5),
Vector(0.5,-0.5,-0.5),
Vector(0.5,-0.5,-0.5)
};

Float BV::normals_t[NUM_PSN]=
{
0,0,0,1,
0.5, -0.5, 0.5, -0.5,
0.5, -0.5, 0.5, -0.5
};


ST_Box World_Box={
-MAX_M,MAX_M,
-MAX_M,MAX_M,
-MAX_M,MAX_M,
-MAX_M,MAX_M
};

inline static bool BV__intersect(BV& bv,Ray& r,Float &T)
{
 // see chapter 1 for how we find the intersection of ray and bv
	Float tnear,tfar;
	Float a,b;
	Float t_near_max,t_far_min;

	// compute tmin and tmax as the first one, then update them as we
	// go along.
	a=(bv.dnear[0]-r.p_dot_b[0])*(r.p_dot_a_rep[0]);
	b=( bv.dfar[0]-r.p_dot_b[0])*(r.p_dot_a_rep[0]);
	if(r.p_dot_a_rep[0]<0){
		t_near_max=b;
		t_far_min=a;
	}
	else{
		t_near_max=a;
		t_far_min=b;
	}
	for(int i=1;i<NUM_PSN;i++){
		// compute tnear,tfar for i
		a=(bv.dnear[i]-r.p_dot_b[i])*(r.p_dot_a_rep[i]);
		b=( bv.dfar[i]-r.p_dot_b[i])*(r.p_dot_a_rep[i]);
		if(r.p_dot_a_rep[i]<0){
			tnear=b;
			tfar =a;
		}
		else{
			tnear=a;
			tfar =b;
		}
		// intersect (tnear,tfar) with running intersection
		if(tnear>t_near_max) t_near_max=tnear;
		if(tfar<t_far_min) t_far_min=tfar;
		// break early if detect
		if(t_near_max>t_far_min) return false;
	}
	// return true if (tnear,tfar) lies infront of ray
	T=t_near_max;
	return t_far_min>=0.0;
}

bool BV::intersect(Ray &r)
{
	Float t;
	return BV__intersect(*this,r,t);
}

bool BV::intersect(Ray &r,Float &t)
{
	return BV__intersect(*this,r,t);
}
void Ray::bv_setup()
{
	// compute P dot a and P dot b for each plane set normal
	// the 4-dot product  (a,b,c,d)*(A,B,C,D)
	//  == (a,b,c)*(A,B,C) + d*D
	// the time component of Ray::d is 0, since it a ray
	// doesn't move through time.
	// the time component of Ray::o is Ray::t
	for(int i=0;i<NUM_PSN;i++){
		Float f=BV::normals[i]*d;
		if(f==0.0) p_dot_a_rep[i]=MAX_M; // from kay&co
		else p_dot_a_rep[i]=1.0/f;
		p_dot_b[i]=BV::normals[i]*o + BV::normals_t[i]*t;
	}
}

class BV_Object: public Object
{
public:
	BV_Object(Object*);
	virtual bool intersect(Ray&,Float&);
	virtual void aux_info(Ray&,Intersect_Info&,Shade_Info&);
	virtual void compute_bounds(BV&);
	virtual bool inside(AA_Box&);
};

BV_Object::BV_Object(Object* o):
	Object(o->surface,o->t0,o->t1,primitive)
{
	bound=o->bound;
}

static bool BV_intersect(BV& bv,Ray& r,Float& T, int& plane)
{
	// takend from BV::intersect
	Float tnear,tfar;
	Float a,b;
	Float t_near_max,t_far_min;

	a=(bv.dnear[0]-r.p_dot_b[0])*(r.p_dot_a_rep[0]);
	b=( bv.dfar[0]-r.p_dot_b[0])*(r.p_dot_a_rep[0]);
	if(r.p_dot_a_rep[0]<0){
		t_near_max=b;
		t_far_min=a;
	}
	else{
		t_near_max=a;
		t_far_min=b;
	}
	plane=0; // plane ray intersects
	for(int i=1;i<NUM_PSN;i++){
		a=(bv.dnear[i]-r.p_dot_b[i])*(r.p_dot_a_rep[i]);
		b=( bv.dfar[i]-r.p_dot_b[i])*(r.p_dot_a_rep[i]);
		if(r.p_dot_a_rep[i]<0){
			tnear=b;
			tfar =a;
		}
		else{
			tnear=a;
			tfar =b;
		}
		// intersect (tnear,tfar) with running intersection
		if(tnear>t_near_max)
			{t_near_max=tnear;plane=i;}
		if(tfar<t_far_min) t_far_min=tfar;
		// break early if detect
		if(t_near_max>t_far_min) return false;
	}
	T=t_near_max;
	return true; // must have intersected
}

bool BV_Object::intersect(Ray& r,Float& T)
{
	int i;
	bool ret =BV_intersect(bound,r,T,i);
	return ret && T>RAY_EPS;
}

void BV_Object::aux_info(Ray& r,Intersect_Info& ii,Shade_Info& si)
{
	int plane; Float T;
	BV_intersect(bound,r,T,plane);
	si.P=r.o+r.d*ii.t;
	if( BV::normals[plane]*r.d <0)
		si.normal=BV::normals[plane];
	else
		si.normal=-BV::normals[plane];
}

void BV_Object::compute_bounds(BV& bv)
{
	bv=bound;
}

bool BV_Object::inside(AA_Box&)
{
	return false; //## this isn't done yet
}


Object* BV_ify(Object* o)
{
	return new BV_Object(o);
}

bool clip(BV& bv,ST_Box& box)
{
	// we know that the first 4 PSNs are the axies
	// check to see if bv is outside box
	if( bv.dnear[0]>box.x1 ||
	    bv.dfar [0]<box.x0 ||
	    bv.dnear[1]>box.y1 ||
	    bv.dfar [1]<box.y0 ||
	    bv.dnear[2]>box.z1 ||
	    bv.dfar [2]<box.z0 ||
	    bv.dnear[3]>box.t1 ||
	    bv.dfar [3]<box.t0)
		return false; // bv outside box
	// OK, only bother clipping axis PSNs, because
	// it doesn't mater if diagonals don't change
	if(bv.dnear[0]<box.x0) bv.dnear[0]=box.x0;
	if(bv.dfar [0]>box.x1) bv.dfar [0]=box.x1;
	if(bv.dnear[1]<box.y0) bv.dnear[1]=box.y0;
	if(bv.dfar [1]>box.y1) bv.dfar [1]=box.y1;
	if(bv.dnear[2]<box.z0) bv.dnear[2]=box.z0;
	if(bv.dfar [2]>box.z1) bv.dfar [2]=box.z1;
	if(bv.dnear[3]<box.t0) bv.dnear[3]=box.t0;
	if(bv.dfar [3]>box.t1) bv.dfar [3]=box.t1;
	return true;
}

BV BV_union(BV& a,BV& b)
{
	BV c=a;
	for(int i=0;i<NUM_PSN;i++){
		if(b.dnear[i]<c.dnear[i]) c.dnear[i]=b.dnear[i];
		if(b.dfar[i]>c.dfar[i]) c.dfar[i]=b.dfar[i];
	}
	return c;
}


