#include "spheres.h"
#include "polygon.h"
#include "transform.h"
#include "ttransform.h"
#include "linear.h"

#define SQR(x) ((x)*(x))

static void minmax(Float &min, Float& max, Float a, Float b, Float c, Float d)
{
	min=a; max=a;
	if(b<min) min=b;
	if(c<min) min=c;
	if(d<min) min=d;
	if(b>max) max=b;
	if(c>max) max=c;
	if(d>max) max=d;
}

static inline void project_point(Float&min,Float&max,L_proj_fn& f, Vector P,Float T0,Float T1)
{
	Float a,b;

	a=eval(f.A,T0)*P.x+eval(f.B,T0)*P.y+eval(f.C,T0)*P.z + eval(f.D,T0);
	b=eval(f.A,T1)*P.x+eval(f.B,T1)*P.y+eval(f.C,T1)*P.z + eval(f.D,T1);
	if(a<b) { min=a; max=b;}
	else { min=b; max=a;}
}

BV compute_bounds(L_TTransformation* LT)
{ // this is taken from TTRansform::compute_bounds()
	BV bv;
	L_proj_fn fn;
	Float t0,t1;
	t0=LT->t0; t1=LT->t1;
	Float seg_width=(t1-t0)/Float(max_segments);

	
	// for each Pi
	for(int i=0;i<NUM_PSN;i++){
		// compute dnear and dfar for Pi
		// by looking at near and far for each line segment
		Float min,max;
		fn= LT->project(0,BV::normals[i],BV::normals_t[i]);
		project_point(min,max,fn,Vector(0,0,0),t0,t0+seg_width);
		for(int seg=1;seg<max_segments;seg++){
			Float a,b;
			fn= LT->project(seg,BV::normals[i],BV::normals_t[i]);
			project_point(a,b,fn,Vector(0,0,0),t0+seg*seg_width,t0+(seg+1)*seg_width);
			if(a<min) min=a;
			if(b>max) max=b;
		}
		bv.dnear[i]=min;
		bv.dfar[i]=max;
	}
	return bv;
}

void Object::compute_bounds(BV& bv)
{
	// for each Pi
	int i;
	L_proj_fn f;
	f.A.k0=0;
	f.B.k0=0;
	f.C.k0=0;
	f.D.k1=0;
	for(i=0;i<NUM_PSN;i++){
		f.A.k1=BV::normals[i].x;
		f.B.k1=BV::normals[i].y;
		f.C.k1=BV::normals[i].z;
		f.D.k0=BV::normals_t[i];
		compute_extent(f,bv.dnear[i],bv.dfar[i],t0,t1);
	}
}


void StaticSphere::compute_extent(L_proj_fn &f,Float &min,Float& max,Float T0,Float T1)
{
	Float lamda;
	Float A,B,C,D;
	Float zero_sqr;
	zero_sqr=SQR(f.A.k1) + SQR(f.B.k1) + SQR(f.C.k1);

	// compute A,B,C and lamda
	if(zero_sqr==0.0){
		A=f.A.k1; B=f.B.k1; C=f.C.k1;
	}
	else{
		Float t;
		t=-(f.A.k0*f.A.k1 + f.B.k0*f.B.k1 + f.C.k0*f.C.k1)/zero_sqr;
		// truncate t to [T0,T1]
		if(t<T0) t=T0;
		if(t>T1) t=T1;
		A=eval(f.A,t); B=eval(f.B,t); C=eval(f.C,t);
	}
	lamda=sqrt(A*A+B*B+C*C)/(2.0*r);
	// compute two values of first=Ax+By+Cz
	Float first1,first2;
	if(lamda==0.0) first1=first2=0.0;
	else{
		Float l=1.0/(2.0*lamda);
		Float x,y,z;
		x=(A*l+p.x); y=(B*l+p.y); z=(C*l+p.z);
		first1=A*x + B*y + C*z;
		l=-l;
		x=(A*l+p.x); y=(B*l+p.y); z=(C*l+p.z);
		first2=A*x + B*y + C*z;
	}
	// compute the four possible extrema e1,e2,e3,e4
	Float e1,e2,e3,e4;
	e1=first1 + eval(f.D,T0);
	e2=first1 + eval(f.D,T1);
	e3=first2 + eval(f.D,T0);
	e4=first2 + eval(f.D,T1);
	minmax(min,max,e1,e2,e3,e4);
}

void Polygon::compute_extent(L_proj_fn& f,Float &min,Float& max,Float T0,Float T1)
{
	// in f()=Ax+By+Cz+D, (x,y,z) are constant (each vertex)
	// for each vertex, compute f(T0), f(T1) and update global
	// min/max

	Float a,b; // min/max for each vert
	project_point(min,max,f,verts[0],t0,t1);
	// for each other vertex
	for(int vert=1;vert<nverts;vert++){
		// project
		project_point(a,b,f,verts[vert],T0,T1);
		// update min/max
		if(a<min) min=a;
		if(b>max) max=b;
	}
}


void  Transformed_Object::compute_bounds(BV& bv)
{
	// this is much the same is the TTransform version, except
	// A,B,C are constant in t
	L_proj_fn fn;
	fn.A.k0=0;
	fn.B.k0=0;
	fn.C.k0=0;

	// for each Pi
	for(int i=0;i<NUM_PSN;i++){
		// compute f()=PA(x,y,z,t) + PT
		// ie, 'project'
		fn.A.k1=BV::normals[i]*Vector(T.trans.r1,T.trans.r4,T.trans.r7);
		fn.B.k1=BV::normals[i]*Vector(T.trans.r2,T.trans.r5,T.trans.r8);
		fn.C.k1=BV::normals[i]*Vector(T.trans.r3,T.trans.r6,T.trans.r9);
		fn.D.k0=BV::normals_t[i];
		fn.D.k1=BV::normals[i]*Vector(T.trans.tx,T.trans.ty,T.trans.tz);
		// compute bounds
		obj->compute_extent(fn,bv.dnear[i],bv.dfar[i],t0,t1);
	}
}

void TTransformed_Object::compute_bounds(BV& bv)
{
	L_TTransformation* LT=L_linearize(T);
	T=LT;
	L_proj_fn fn;
	Float seg_width=(t1-t0)/Float(max_segments);

	// for each Pi
	for(int i=0;i<NUM_PSN;i++){
		// compute dnear and dfar for Pi
		// by looking at near and far for each line segment
		Float min,max;
		fn= LT->project(0,BV::normals[i],BV::normals_t[i]);
		obj->compute_extent(fn,min,max,t0,t0+seg_width);
		for(int seg=1;seg<max_segments;seg++){ //## change this "from index of T0 to index of T1
			Float a,b;
			fn= LT->project(seg,BV::normals[i],BV::normals_t[i]);
			obj->compute_extent(fn,a,b,t0+seg*seg_width,t0+(seg+1)*seg_width);
			if(a<min) min=a;
			if(b>max) max=b;
		}
		bv.dnear[i]=min;
		bv.dfar[i]=max;
	}
}

