#include "spheres.h"
#include "polygon.h"
#include "ttransform.h"
#include "sourcegen.h"
#include "camera.h"
#include "light.h"

#include <string.h>
#include <stdio.h>

/* code for generating C code for things */

char* SI_generate_code(Surface_Info si)
{
	char s[256];
	sprintf(s,"\
Surface_Info si;\n\
si.kdr=%.14lg;\n\
si.ksr=%.14lg;\n\
si.kst=%.14lg;\n\
si.n=%.14lg;\n\
si.rcolour=Colour(%.14lg,%.14lg,%.14lg);\n\
si.tcolour=Colour(%.14lg,%.14lg,%.14lg);\n\
si.dcolour=Colour(%.14lg,%.14lg,%.14lg);\n\
si.ior=%.14lg;\n",
si.kdr,si.ksr,si.kst,si.n,si.rcolour.r,si.rcolour.g,si.rcolour.b,
si.tcolour.r,si.tcolour.g,si.tcolour.b,
si.dcolour.r,si.dcolour.g,si.dcolour.b,
si.ior);
	return STR_dup(s);
}

void StaticSphere::generate_code(Source_Generator* sg, bool Root)
{
	root=Root;
	char sph[256];
	char buf[SG_ID_LEN];
	sprintf(sph,"\
%s=Create_Static_Sphere(Vector(%.14lg,%.14lg,%.14lg),%.14lg,si);}\n",
		Source_Generator::name_of(this,buf),
		double(p.x),double(p.y),double(p.z),double(r));

	char *s;
	s=STR_add("{",SI_generate_code(surface));
	s=STR_add(s,sph);
	sg->add_object_code(this,s);
}

Polygon::Polygon(Surface_Info&si,int Nverts, UV* aUVs, int aDA, Vector P, Float d):
	Object(si,primitive),Pn(P),D(d),nverts(Nverts),DA(aDA)
{
	uvs=aUVs;// this is assumed to be a static array in phase_2
	uvps=new UV[nverts];
}


void Polygon::generate_code(Source_Generator* sg, bool Root)
{
	root=Root;
	char *Vs=new char[60*nverts];
	*Vs=0;
	char a[256];
	char buf[SG_ID_LEN];
	int i;
	strcat(Vs,"{static Polygon::UV uvs[]={");
	for(i=0;i<nverts;i++){
		sprintf(a,"%.14lg,%.14lg,",double(uvs[i].u),double(uvs[i].v));
		strcat(Vs,a);
	}
	strcat(Vs,"};\n");
	char *s;
	s=STR_add(Vs,SI_generate_code(surface));
	sprintf(a,"\
%s=new Polygon(si,%d,uvs,%d,Vector(%.14lg,%.14lg,%.14lg),%.14lg);}\n",
Source_Generator::name_of(this,buf),nverts,DA,
double(Pn.x),double(Pn.y),double(Pn.z),D);
	s=STR_add(s,a);
	sg->add_object_code(this,s);
}

void Transformed_Object::generate_code(Source_Generator* sg, bool Root)
{
	root=Root;
	T.generate_code(sg);
	char s[256];
	char buf1[SG_ID_LEN];
	char buf2[SG_ID_LEN];
	char buf3[SG_ID_LEN];
	sprintf(s,"\
%s=Transform_Object(%s,%s);\n",
Source_Generator::name_of(this,buf1),
Source_Generator::name_of(obj,buf2),
Source_Generator::name_of(&T,buf3));
	sg->add_object_code(this,s);
	obj->generate_code(sg);
}

Transformation::Transformation(Float* t,Float* i)
{
	trans.r1=t[0]; trans.r2=t[1]; trans.r3=t[2];
	trans.r4=t[3]; trans.r5=t[4]; trans.r6=t[5];
	trans.r7=t[6]; trans.r8=t[7]; trans.r9=t[8];
	trans.tx=t[9]; trans.ty=t[10]; trans.tz=t[11];

	inverse.r1=i[0]; inverse.r2=i[1]; inverse.r3=i[2];
	inverse.r4=i[3]; inverse.r5=i[4]; inverse.r6=i[5];
	inverse.r7=i[6]; inverse.r8=i[7]; inverse.r9=i[8];
	inverse.tx=i[9]; inverse.ty=i[10]; inverse.tz=i[11];
}


void Transformation::generate_code(Source_Generator* sg)
{
	char a[512];
	char b[SG_ID_LEN];
	char *s;
	strcpy(a,"{static Float trans[]={");
	sprintf(b,"%.14lg,",double(trans.r1)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r2)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r3)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r4)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r5)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r6)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r7)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r8)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.r9)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.tx)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.ty)); strcat(a,b);
	sprintf(b,"%.14lg,",double(trans.tz)); strcat(a,b);
	strcat(a,"};\n");
	s=STR_dup(a);
	strcpy(a,"static Float inverse[]={");
	sprintf(b,"%.14lg,",double(inverse.r1)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r2)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r3)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r4)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r5)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r6)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r7)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r8)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.r9)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.tx)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.ty)); strcat(a,b);
	sprintf(b,"%.14lg,",double(inverse.tz)); strcat(a,b);
	strcat(a,"};\n");
	s=STR_add(s,a);
	sprintf(a,"%s=new Transformation(trans,inverse);}\n",
		Source_Generator::name_of(this,b));
	sg->add_transform_code(this,STR_add(s,a));
}

TTransformation::TTransformation(Function** t,Function** i):
	last_t(-MAX_TIME)
{
	Ttrans.r1=t[0]; Ttrans.r2=t[1]; Ttrans.r3=t[2];
	Ttrans.r4=t[3]; Ttrans.r5=t[4]; Ttrans.r6=t[5];
	Ttrans.r7=t[6]; Ttrans.r8=t[7]; Ttrans.r9=t[8];
	Ttrans.tx=t[9]; Ttrans.ty=t[10]; Ttrans.tz=t[11];

	Tinverse.r1=i[0]; Tinverse.r2=i[1]; Tinverse.r3=i[2];
	Tinverse.r4=i[3]; Tinverse.r5=i[4]; Tinverse.r6=i[5];
	Tinverse.r7=i[6]; Tinverse.r8=i[7]; Tinverse.r9=i[8];
	Tinverse.tx=i[9]; Tinverse.ty=i[10]; Tinverse.tz=i[11];
}

void TTransformed_Object::generate_code(Source_Generator* sg, bool Root)
{
	root=Root;
	T.generate_code(sg);
	char s[256];
	char buf1[SG_ID_LEN];
	char buf2[SG_ID_LEN];
	char buf3[SG_ID_LEN];
	sprintf(s,"\
%s=Transform_Object(%s,%s);\n",
Source_Generator::name_of(this,buf1),
Source_Generator::name_of(obj,buf2),
Source_Generator::name_of(&T,buf3));
	sg->add_object_code(this,s);
	obj->generate_code(sg);
}

void TTransformation::generate_code(Source_Generator* sg)
{
	char a[512];
	char b[SG_ID_LEN];
	char buf[SG_ID_LEN];
	char *s;
	strcpy(a,"{static Function* trans[]={");
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r1,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r2,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r3,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r4,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r5,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r6,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r7,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r8,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.r9,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.tx,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.ty,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Ttrans.tz,buf)); strcat(a,b);
	strcat(a,"};\n");
	s=STR_dup(a);
	strcpy(a,"static Function* inverse[]={");
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r1,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r2,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r3,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r4,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r5,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r6,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r7,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r8,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.r9,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.tx,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.ty,buf)); strcat(a,b);
	sprintf(b,"%s,",Source_Generator::name_of(Tinverse.tz,buf)); strcat(a,b);
	strcat(a,"};\n");
	s=STR_add(s,a);
	sprintf(a,"%s=new TTransformation(trans,inverse);}\n",
		Source_Generator::name_of(this,b));
	sg->add_ttransform_code(this,STR_add(s,a));
	Ttrans.r1->generate_code(sg); Ttrans.r2->generate_code(sg);
	Ttrans.r3->generate_code(sg); Ttrans.r4->generate_code(sg);
	Ttrans.r5->generate_code(sg); Ttrans.r6->generate_code(sg);
	Ttrans.r7->generate_code(sg); Ttrans.r8->generate_code(sg);
	Ttrans.r9->generate_code(sg); Ttrans.tx->generate_code(sg);
	Ttrans.ty->generate_code(sg); Ttrans.tz->generate_code(sg);

	Tinverse.r1->generate_code(sg); Tinverse.r2->generate_code(sg);
	Tinverse.r3->generate_code(sg); Tinverse.r4->generate_code(sg);
	Tinverse.r5->generate_code(sg); Tinverse.r6->generate_code(sg);
	Tinverse.r7->generate_code(sg); Tinverse.r8->generate_code(sg);
	Tinverse.r9->generate_code(sg); Tinverse.tx->generate_code(sg);
	Tinverse.ty->generate_code(sg); Tinverse.tz->generate_code(sg);
}

// camera: use 'cam' as name of variable
void Camera::generate_code(Source_Generator* sg)
{
	char s[14*40]; // should be big enough
	char buf01[SG_ID_LEN];
	char buf02[SG_ID_LEN];
	char buf03[SG_ID_LEN];
	char buf04[SG_ID_LEN];
	char buf05[SG_ID_LEN];
	char buf06[SG_ID_LEN];
	char buf07[SG_ID_LEN];
	char buf08[SG_ID_LEN];
	char buf09[SG_ID_LEN];
	char buf10[SG_ID_LEN];
	char buf11[SG_ID_LEN];
	char buf12[SG_ID_LEN];
	sprintf(s,"\
cam.width=%d;\n\
cam.height=%d;\n\
cam.fname=\"%s\";\n\
cam.C=TVector(%s,%s,%s);\n\
cam.look_at=TVector(%s,%s,%s);\n\
cam.up=TVector(%s,%s,%s);\n\
cam.f=%s;\n\
cam.xs=%s;\n\
cam.ys=%s;\n\
cam.backdrop=Colour(%.14lg,%.14lg,%.14lg);\n\
cam.end_time=%.14lg;\n\
cam.framesec=%.14lg;\n\
cam.sub=%d;\n\
cam.motion_blur=%d;\n\
cam.anti_alias=%d;\n\
",
width,height,fname,
Source_Generator::name_of(C.xfn,buf01),Source_Generator::name_of(C.yfn,buf02),Source_Generator::name_of(C.zfn,buf03),
Source_Generator::name_of(look_at.xfn,buf04),Source_Generator::name_of(look_at.yfn,buf05),Source_Generator::name_of(look_at.zfn,buf06),
Source_Generator::name_of(up.xfn,buf07),Source_Generator::name_of(up.yfn,buf08),Source_Generator::name_of(up.zfn,buf09),
Source_Generator::name_of(f,buf10),
Source_Generator::name_of(xs,buf11),
Source_Generator::name_of(ys,buf12),
double(backdrop.r),double(backdrop.g),double(backdrop.b),
double(end_time),double(framesec),
sub,motion_blur,anti_alias);

	C.xfn->generate_code(sg);C.yfn->generate_code(sg);C.zfn->generate_code(sg);
	look_at.xfn->generate_code(sg);look_at.yfn->generate_code(sg);look_at.zfn->generate_code(sg);
	up.xfn->generate_code(sg);up.yfn->generate_code(sg);up.zfn->generate_code(sg);
	f->generate_code(sg);
	xs->generate_code(sg);
	ys->generate_code(sg);

	sg->add_camera_code(this,s);
}

// for functions

void Function::generate_code(Source_Generator* sg)
{
	char *s;
	s=C_string_rep(sg);
	s=STR_add("return ",s);
	s=STR_add(s,";\n");
	sg->add_function_code(this,s);
}

char* FN_constant::C_string_rep(Source_Generator*)
{
	char s[32];
	sprintf(s,"%.14lg",double(c));
	return STR_dup(s);
}


char* FN_linear::C_string_rep(Source_Generator*)
{
	char s[20*2+10];
	sprintf(s,"(%.14lg*t)+(%.14lg)",double(a),double(b));
	return STR_dup(s);
}

char* FN_cos_fn::C_string_rep(Source_Generator* sg)
{
	char *S[3];
	S[0]="cos(";
	S[1]=fn->C_string_rep(sg);
	S[2]=")";
	return STR_add(3,S);
}

char* FN_sin_fn::C_string_rep(Source_Generator*sg)
{
	char *S[3];
	S[0]="sin(";
	S[1]=fn->C_string_rep(sg);
	S[2]=")";
	return STR_add(3,S);
}

char* FN_neg_sin_fn::C_string_rep(Source_Generator*sg)
{
	char *S[3];
	S[0]="(-sin(";
	S[1]=fn->C_string_rep(sg);
	S[2]="))";
	return STR_add(3,S);
}

char* FN_cos::C_string_rep(Source_Generator*sg)
{
	char s[256];
	sprintf(s,"(%.14lg * cos(%.14lg * t+%.14lg)+(%.14lg))",
		double(m),double(a),double(b),double(c));
	return STR_dup(s);
}

char* FN_sin::C_string_rep(Source_Generator*)
{
	char s[256];
	sprintf(s,"(%.14lg * sin(%.14lg * t+%.14lg)+(%.14lg))",
		double(m),double(a),double(b),double(c));
	return STR_dup(s);
}

void FN_composite::generate_code(Source_Generator* sg)
{
	char *buf=new char[n*60];
	char id_buf[SG_ID_LEN];
	int i;
	strcpy(buf,"static Function* fns[]={");
	for(i=0;i<n;i++){
		strcat(buf,Source_Generator::name_of(fns[i],id_buf));
		strcat(buf,",");
		fns[i]->generate_code(sg);
	}
	strcat(buf,"};\n");
	char buf2[200];
sprintf(buf2,"\
Function* o;\n\
o=fns[%d];\n\
for(int i=0;i<%d;i++)\n\
if(fns[i]->t0<=t && t<fns[i]->t1){\n\
o=fns[i];\n\
break;\n\
}\n\
return o->eval(t);\n",n-1,n-1);
	sg->add_function_code(this,STR_add(buf,buf2));
}


char* FN_composite::C_string_rep(Source_Generator* sg)
{
	char buf[SG_ID_LEN+20];
	Source_Generator::name_of(this,buf);
	strcat(buf,"->eval(t)");
	generate_code(sg);
	return STR_dup(buf);
}

char* FN_multiply::C_string_rep(Source_Generator* sg)
{
	char *S[5];
	S[0]="(";
	S[1]=a->C_string_rep(sg);
	S[2]=")*(";
	S[3]=b->C_string_rep(sg);
	S[4]=")";
	return STR_add(5,S);
}

char* FN_add::C_string_rep(Source_Generator* sg)
{
	char* S[5];
	S[0]="(";
	S[1]=f1->C_string_rep(sg);
	S[2]=")+(";
	S[3]=f2->C_string_rep(sg);
	S[4]=")";

	return STR_add(5,S);
}

char* FN_neg::C_string_rep(Source_Generator* sg)
{
	char *s=STR_add("(-(",fn->C_string_rep(sg));
	return STR_add(s,"))");
}

char* FN_recip::C_string_rep(Source_Generator*sg)
{
	char *s=STR_add("1.0/(",fn->C_string_rep(sg));
	return STR_add(s,")");
}


void Light_generate_code(Light *l,Source_Generator* sg)
{
	char s[256];
	char buf1[SG_ID_LEN];
	char buf2[SG_ID_LEN];
	char buf3[SG_ID_LEN];
	sprintf(s,"\
K_add_light(new Light(TVector(%s,%s,%s),Colour(%.14lg,%.14lg,%.14lg)));\n",
Source_Generator::name_of(l->pos.xfn,buf1),
Source_Generator::name_of(l->pos.yfn,buf2),
Source_Generator::name_of(l->pos.zfn,buf3),
double(l->colour.r),
double(l->colour.g),
double(l->colour.b));
	l->pos.xfn->generate_code(sg);
	l->pos.yfn->generate_code(sg);
	l->pos.zfn->generate_code(sg);
	sg->add_light_code(l,s);
}

