#ifndef __LINEAR_H
#define __LINEAR_H

#include "config.h"
#include "ttransform.h"

extern int max_segments;
struct L_coeffs{
	Float k0,k1;
};

inline Float eval(L_coeffs l,Float t){
	return t*l.k0+l.k1;
}

struct L_proj_fn{ // projected function f(x,y,z,t)=Ax+By+Cz+D
	L_coeffs A,B,C,D;
};

// L_fn represents a piecwise linear curve
struct L_fn{
        L_coeffs* c;
	Float T0,T1;
};

struct L_Matrix{ L_fn f[12];};

void L_init(int nframes);

class L_TTransformation: public TTransformation{
public:
	L_TTransformation(Float,Float);
	L_proj_fn project(int i,Vector P, Float Pt); // ith time interval
	virtual void eval_inverse(Float t);
	virtual bool is_linear(){return true;}
	Vector apply(Float t); // applies transformation to point(0,0,0)
	L_Matrix M;
};

L_TTransformation* L_linearize(TTransformation*);

// computes the bounds of the transformation applied to the point (0,0,0)
BV compute_bounds(L_TTransformation*);

#endif

