
#include "polygon.h"


Polygon::Polygon(Surface_Info& si, int Nverts, Vector* Verts,Float T0,Float T1):
	Object(si,T0,T1,primitive),nverts(Nverts)
{
	int i;

	// make a local copy of verts
	verts=new Vector[nverts];
	for(i=0;i<nverts;i++)
		verts[i]=Verts[i];

	// making copy of (projected) verticies
	uvs=new UV[nverts];
	uvps=new UV[nverts];

	// find Pn and D
	Vector a,b;
	a=verts[0]-verts[1];
	b=verts[2]-verts[1];
	Pn=cross(b,a);
	Pn.normalize();
	D=-(Pn*verts[0]);
	
	// compute uvs, so determin dominant axis
	Float A,B,C;
	A=ABS(Pn.x);
	B=ABS(Pn.y);
	C=ABS(Pn.z);
	if(A>B && A>C){ // X is dominant axis
		DA=0;
		for(i=0;i<nverts;i++){
			uvs[i].u=verts[i].y;
			uvs[i].v=verts[i].z;
		}
	}
	else if(B>A && B>C){ // Y is dominant axis
		DA=1;
		for(i=0;i<nverts;i++){
			uvs[i].u=verts[i].x;
			uvs[i].v=verts[i].z;
		}
	}
	else{ // Z is dominant axis
		DA=2;
		for(i=0;i<nverts;i++){
			uvs[i].u=verts[i].x;
			uvs[i].v=verts[i].y;
		}
	}
}

bool Polygon::intersect(Ray& r,Float& T)
{
	// this uses haines' technique from Intro. to Ray Tracing

	// find the intersection with the plane
	Float t,vd,vo;
	vd=Pn*r.d;
	if(vd<RAY_EPS && vd>(-RAY_EPS)) return false; // parallel to plane
	vo=-(Pn*r.o +D);
	t=vo/vd;
	if(t<RAY_EPS) return false; // plane behind ray origin

	T=t;

	// calculate intersection point
	Vector P=r.o+r.d*t;
	// project P onto u,v
	Float u,v;
	if(DA==0) { u=P.y; v=P.z;}
	else if(DA==1) { u=P.x; v=P.z;}
	else { u=P.x; v=P.y;}

	// calculate uvps
	int i;
	for(i=0;i<nverts;i++){
		uvps[i].u=uvs[i].u-u;
		uvps[i].v=uvs[i].v-v;
	}
	int NC=0,SH,NSH;
	int a,b;
	if(uvps[0].v<0) SH=-1;
	else SH=1;

	// test againts each edge
	for(a=0,b=1;a<nverts;a++,b=(a+1)%nverts){
		if(uvps[b].v<0) NSH=-1;
		else NSH=1;
		if(SH != NSH){
			if(uvps[a].u>0 && uvps[b].u>0) NC++;
			else if(uvps[a].u>0 || uvps[b].u>0){
				// if Ua - Va*(Ub-Ua)/(Vb-Va) > 0
				if( (uvps[a].u- uvps[a].v*
				 (uvps[b].u-uvps[a].u)/(uvps[b].v-uvps[a].v))>0)
					NC++;
			}
			SH=NSH;
		}
	}
	// if NC is odd, point is inside polygon
	if(NC&1) return true;
	return false;
}

void Polygon::aux_info(Ray& r,Intersect_Info& ii,Shade_Info &si)
{
	si.P=r.o+r.d*ii.t;
	if( Pn*r.d <0)
		si.normal=Pn;
	else
		si.normal=-Pn;
}


Object* Create_Polygon( int nverts,Vector* verts,Float t0,Float t1,Surface_Info si)
{
	return new Polygon(si, nverts, verts,t0,t1);
}


