#include "shader.h"
#include "colour.h"
#include "light.h"
#include "object.h"
#include <math.h>

// a kinda whitted shader

// basically
//  I= Cr [ kdr sum(Ilj N.Lj) + ksr Isr] + Ct [ kst Ist ]

static Colour ambient_light;
static Colour bounding_colour;
void S_init(Colour ambient,Colour sky)
{
	ambient_light=ambient;
	bounding_colour=sky;
}

void S_shade(Ray_Info& ri, Intersect_Info& ii, Colour& RetColour)
{
	Shade_Info si;
	Colour colour(0,0,0);
	bool do_reflection=true;

	// set up some vectors we are going to need
	Ray L; // ray to light source
	// we have normal in si.normal
	ii.obj->aux_info(ri.ray,ii,si);
	Vector P=si.P; // point of ray intersection

	// compute contribution of light sources
	Light* ptr=K_get_lights();// list of lights
	L.o=P;
	L.t=ri.ray.t;
	Float f; // distance to light
	Float s;
	Colour tcol(0,0,0); // temp colour

	while(ptr){ // for each light source
		// compute direction to light
		L.d= ptr->pos(ri.ray.t)-P;
		f=length(L.d);
		L.d=L.d*(1/f); // normalize direction
		s=si.normal*L.d;
		L.bv_setup();
		if(s>RAY_EPS && !K_shadow_test(L,f,ptr)){ // we have no shadow
		  Colour c=ptr->colour*pow(s,ii.obj->surface.n); // falloff?
		  tcol+=c;
		}
		ptr=ptr->next;
	}
	if(! ri.object_in) tcol+=ambient_light; // outside an object
	tcol*=ii.obj->surface.kdr;
	tcol*=ii.obj->surface.dcolour; // colour reflectence curve
	colour+=tcol;
	// compute specular parts
	Ray_Info RI=ri;

	RI.generation++;

	Float IN = si.normal * ri.ray.d; // I dot N


	// specular transmission
	if(ii.obj->surface.ior!=0.0){ // not opaque
		// compute transmition direction, from _intro2RT_ pg140
		Float n1,n2,nit,Ci,d;
		// n1 is object of object we are in
		// n2 is object we are entering
		n1= ri.object_in ? ri.object_in->surface.ior : 1.0;
		if(ii.obj==ri.object_in) // we are leaving ourselves
			n2=1.0;
		else n2= ii.obj->surface.ior;
		nit=n1/n2; // rel. index of refraction
		Ci=-IN;
		d=1+nit*nit*(Ci*Ci-1);
		// if already inside an object, don't
		// spawn reflection ray by default
		if(ri.object_in) do_reflection=false;
		if(d>=0.0){ // TIR did not occur
			RI.ray.o=P;
			RI.ray.d= ri.ray.d*nit + si.normal*(nit*Ci-sqrt(d));
			RI.ray.d.normalize();
			RI.weight=ri.weight*ii.obj->surface.kst;
			RI.type=Ray_Info::transmitted;
			RI.object_in=ii.obj;
			tcol=Colour(0,0,0);
			RI.ray.bv_setup();
			if(K_trace(RI,tcol)) ;
			else  // bounding colour
				tcol=bounding_colour;
			tcol*=ii.obj->surface.kst;
			tcol*=ii.obj->surface.tcolour;
			colour+=tcol;
		}
		// only do reflection inside an object if TIR occurs
		else if(ri.object_in)
			do_reflection=true;// don't do 
	}
	// specular reflection
	if(do_reflection){
		// compute new ray  R=I -2(N.I)N
		RI.ray.o=P;
		RI.ray.d= ri.ray.d +si.normal * ( -2.0*IN );
		RI.type=Ray_Info::reflected;
		RI.weight*=ii.obj->surface.ksr;
		tcol=Colour(0,0,0);
		RI.ray.bv_setup();
		if(K_trace(RI,tcol)) ;
		else tcol=bounding_colour;
		tcol*=ii.obj->surface.ksr;
		tcol*=ii.obj->surface.rcolour;
		colour+=tcol;
	}

	RetColour+=colour;
	}


