#include "spheres.h"


StaticSphere::StaticSphere(Surface_Info& si,Vector P,Float R,Float T0,Float T1):
	Object(si,T0,T1,primitive), p(P),r(R)
{
	r2=r*r;
}

bool StaticSphere::intersect(Ray& ray, Float& t)
{
	// from _intro_to_raytracing_ p36
	Float A,B,C;
	Vector V;
	V=ray.o-p;
	A=ray.d*ray.d;
	B=2*( ray.d * V );
	C=V*V - r2;
	Float d;
	d=B*B-4*A*C;
	if(d<=0.0) return false;
	d=sqrt(d);
	t=(-B - d) /(2*A);
	if(t>=RAY_EPS) return true;
	t=(-B+d)/(2*A);
	if(t>=RAY_EPS) return true;
	return false;
}

void StaticSphere::aux_info(Ray& ray,Intersect_Info& ii,Shade_Info& si)
{
	si.P=ray.o+ray.d*ii.t;
	si.normal=(si.P-p)*(1/r);
	// if ray origin in inside sphere, reverse normal direction
	if(si.normal * ray.d > 0) si.normal=-si.normal;
}


Object* Create_Static_Sphere(Vector P,Float r,Float t0,Float t1,Surface_Info si)
{
	return new StaticSphere(si,P,r,t0,t1);
}

