#include "ttransform.h"
#include "object.h"
#include <math.h>

TMatrix::TMatrix(Float T0,Float T1):
	t0(T0),t1(T1)
{
	r1=new FN_constant(1,t0,t1);
	r2=new FN_constant(0,t0,t1);
	r3=new FN_constant(0,t0,t1);
	r4=new FN_constant(0,t0,t1);
	r5=new FN_constant(1,t0,t1);
	r6=new FN_constant(0,t0,t1);
	r7=new FN_constant(0,t0,t1);
	r8=new FN_constant(0,t0,t1);
	r9=new FN_constant(1,t0,t1);
	tx=new FN_constant(0,t0,t1);
	ty=new FN_constant(0,t0,t1);
	tz=new FN_constant(0,t0,t1);
}

Matrix TMatrix::operator()(Float t)
{
	Matrix a;
	a.r1=r1->eval(t); a.r2=r2->eval(t); a.r3=r3->eval(t); a.tx=tx->eval(t);
	a.r4=r4->eval(t); a.r5=r5->eval(t); a.r6=r6->eval(t); a.ty=ty->eval(t);
	a.r7=r7->eval(t); a.r8=r8->eval(t); a.r9=r9->eval(t); a.tz=tz->eval(t);
	return a;
}

TMatrix operator* (TMatrix& a, TMatrix& b)
{
/* diagram to help with computing this matrix multiply
[ c1 c2 c3 cx ]   [ a1 a2 a3 ax ] [ b1 b2 b3 bx ]
[ c4 c5 c6 cy ]   [ a4 a5 a6 ay ] [ b4 b5 b6 by ]
[ c7 c8 c9 cz ] = [ a7 a8 a9 az ].[ b7 b8 b9 bz ]
[ 0  0  0  1  ]   [ 0  0  0  1  ] [ 0  0  0  1  ]  */
	TMatrix c(a.t0,a.t1);
	c.r1=add(mul(a.r1,b.r1) , mul(a.r2,b.r4) , mul(a.r3,b.r7));
	c.r2=add(mul(a.r1,b.r2) , mul(a.r2,b.r5) , mul(a.r3,b.r8));
	c.r3=add(mul(a.r1,b.r3) , mul(a.r2,b.r6) , mul(a.r3,b.r9));

	c.r4=add(mul(a.r4,b.r1) , mul(a.r5,b.r4) , mul(a.r6,b.r7));
	c.r5=add(mul(a.r4,b.r2) , mul(a.r5,b.r5) , mul(a.r6,b.r8));
	c.r6=add(mul(a.r4,b.r3) , mul(a.r5,b.r6) , mul(a.r6,b.r9));

	c.r7=add(mul(a.r7,b.r1) , mul(a.r8,b.r4) , mul(a.r9,b.r7));
	c.r8=add(mul(a.r7,b.r2) , mul(a.r8,b.r5) , mul(a.r9,b.r8));
	c.r9=add(mul(a.r7,b.r3) , mul(a.r8,b.r6) , mul(a.r9,b.r9));

	c.tx=add(mul(a.r1,b.tx) , mul(a.r2,b.ty) , mul(a.r3,b.tz) ,a.tx);
	c.ty=add(mul(a.r4,b.tx) , mul(a.r5,b.ty) , mul(a.r6,b.tz) ,a.ty);
	c.tz=add(mul(a.r7,b.tx) , mul(a.r8,b.ty) , mul(a.r9,b.tz) ,a.tz);

	return c;
}

// Transformations



void TTransformation::rotateX(Function* fn)
{
	TMatrix M(t0,t1);
	FN_cos_fn *c=new FN_cos_fn(fn,t0,t1);
	FN_sin_fn *s1=new FN_sin_fn(fn,t0,t1);
	FN_neg_sin_fn *s2=new FN_neg_sin_fn(fn,t0,t1);
	M.r5=c; M.r6=s2;
	M.r8=s1; M.r9=c;
	Ttrans=M*Ttrans;
	M.r6=s1; M.r8=s2;
	Tinverse=Tinverse*M;
}

void TTransformation::rotateY(Function* fn)
{
	TMatrix M(t0,t1);
	FN_cos_fn *c=new FN_cos_fn(fn,t0,t1);
	FN_sin_fn *s1=new FN_sin_fn(fn,t0,t1);
	FN_neg_sin_fn *s2=new FN_neg_sin_fn(fn,t0,t1);
	M.r1=c; M.r3=s1;
	M.r7=s2; M.r9=c;
	Ttrans=M*Ttrans;
	M.r3=s2; M.r7=s1;
	Tinverse=Tinverse*M;
}

void TTransformation::rotateZ(Function* fn)
{
	TMatrix M(t0,t1);
	FN_cos_fn *c=new FN_cos_fn(fn,t0,t1);
	FN_sin_fn *s1=new FN_sin_fn(fn,t0,t1);
	FN_neg_sin_fn *s2=new FN_neg_sin_fn(fn,t0,t1);
	M.r1=c; M.r2=s2;
	M.r4=s1; M.r5=c;
	Ttrans=M*Ttrans;
	M.r2=s1; M.r4=s2;
	Tinverse=Tinverse*M;
}

void TTransformation::translate(TVector T)
{
	TMatrix M(t0,t1);
	M.tx=T.xfn; M.ty=T.yfn; M.tz=T.zfn;
	Ttrans=M*Ttrans;
	M.tx=neg(T.xfn); M.ty=neg(T.yfn); M.tz=neg(T.zfn);
	Tinverse=Tinverse*M;
}

void TTransformation::scale(TVector T)
{
	TMatrix M(t0,t1);
	M.r1=T.xfn; M.r5=T.yfn; M.r9=T.zfn;
	Ttrans=M*Ttrans;
	M.r1=reciprical(T.xfn); M.r5=reciprical(T.yfn); M.r9=reciprical(T.zfn);
	Tinverse=Tinverse*M;
}

void TTransformation::post_mult(TTransformation* T)
{
	Ttrans=Ttrans*T->Ttrans;
	Tinverse=T->Tinverse* Tinverse;
}

void TTransformation::eval_inverse(Float t)
{
	if(t==last_t) return;
	last_t=t;
	inverse=Tinverse(t);
}

TTransformation::TTransformation(Float T0,Float T1):
	Ttrans(T0,T1),Tinverse(T0,T1),last_t(-MAX_TIME),t0(T0),t1(T1) {};

TTransformation::TTransformation(Transformation* T,Float T0,Float T1):
	Ttrans(T0,T1),Tinverse(T0,T1),last_t(-MAX_TIME),t0(T0),t1(T1)
{
	Ttrans.r1=new FN_constant(T->trans.r1,t0,t1);
	Ttrans.r2=new FN_constant(T->trans.r2,t0,t1);
	Ttrans.r3=new FN_constant(T->trans.r3,t0,t1);
	Ttrans.r4=new FN_constant(T->trans.r4,t0,t1);
	Ttrans.r5=new FN_constant(T->trans.r5,t0,t1);
	Ttrans.r6=new FN_constant(T->trans.r6,t0,t1);
	Ttrans.r7=new FN_constant(T->trans.r7,t0,t1);
	Ttrans.r8=new FN_constant(T->trans.r8,t0,t1);
	Ttrans.r9=new FN_constant(T->trans.r9,t0,t1);
	Ttrans.tx=new FN_constant(T->trans.tx,t0,t1);
	Ttrans.ty=new FN_constant(T->trans.ty,t0,t1);
	Ttrans.tz=new FN_constant(T->trans.tz,t0,t1);

	Tinverse.r1=new FN_constant(T->inverse.r1,t0,t1);
	Tinverse.r2=new FN_constant(T->inverse.r2,t0,t1);
	Tinverse.r3=new FN_constant(T->inverse.r3,t0,t1);
	Tinverse.r4=new FN_constant(T->inverse.r4,t0,t1);
	Tinverse.r5=new FN_constant(T->inverse.r5,t0,t1);
	Tinverse.r6=new FN_constant(T->inverse.r6,t0,t1);
	Tinverse.r7=new FN_constant(T->inverse.r7,t0,t1);
	Tinverse.r8=new FN_constant(T->inverse.r8,t0,t1);
	Tinverse.r9=new FN_constant(T->inverse.r9,t0,t1);
	Tinverse.tx=new FN_constant(T->inverse.tx,t0,t1);
	Tinverse.ty=new FN_constant(T->inverse.ty,t0,t1);
	Tinverse.tz=new FN_constant(T->inverse.tz,t0,t1);
}

TTransformed_Object::TTransformed_Object(Object* o,TTransformation* t,Surface_Info &s):
	Object(s,t->t0,t->t1,ttransform),obj(o),T(t)
{}

bool TTransformed_Object::intersect(Ray& r,Float& t)
{
	Ray ray=r;
	T->eval_inverse(r.t);
	ray.o=transform_point(T->inverse,r.o);
	ray.d=transform_normal(T->inverse,r.d);
	return obj->intersect(ray,t);
}

void TTransformed_Object::aux_info(Ray& r,Intersect_Info& ii,Shade_Info& si)
{
	Ray ray;
	T->eval_inverse(r.t);
	ray.o=transform_point(T->inverse,r.o);
	ray.d=transform_normal(T->inverse,r.d);
	obj->aux_info(ray,ii,si);
	// transform normal and point to world co-ords
	//si.P=transform_point(T.trans,si.P);
	si.P=r.o+r.d*ii.t;
	si.normal=transform_normalT(T->inverse,si.normal);
	si.normal.normalize();
}


Object* Transform_Object(Object* obj,TTransformation* T,Surface_Info& si)
{
	if(obj->type==Object::primitive)
		return new TTransformed_Object(obj,T,si);
	TTransformation* S;
	Object* prim_obj;
	if(obj->type==Object::transform){
		S=new TTransformation( &((Transformed_Object*)(obj))->T,T->t0,T->t1);
		prim_obj= ((Transformed_Object*)(obj))->obj;
	}
	if(obj->type==Object::ttransform){
		S=((TTransformed_Object*)(obj))->T;
		prim_obj= ((TTransformed_Object*)(obj))->obj;
	}
	TTransformation *T1=new TTransformation(*T);
	T1->post_mult(S);
	return new TTransformed_Object(prim_obj,T1,si);
}


